package de.ibh.StrongRoom.Client;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.nio.charset.Charset;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.bouncycastle.openpgp.PGPException;
import org.bouncycastle.openpgp.PGPPublicKeyRing;
import org.bouncycastle.openpgp.PGPSecretKeyRing;
import org.bouncycastle.openpgp.PGPUtil;
import org.media_art_online.gnupg.GnuPG;

import de.ibh.StrongRoom.Client.Config.BaseConf;
import de.ibh.StrongRoom.Client.Config.SigningConf;
import de.ibh.StrongRoom.Client.Config.BaseConf.ESecretKeySource;
import de.ibh.StrongRoom.Crypto.CryptoBackend;
import de.ibh.StrongRoom.Crypto.IPassphraseRequest;
import de.ibh.StrongRoom.Crypto.IRetrievePublicKeyRingHelper;
import de.ibh.StrongRoom.Crypto.IRetrieveSecretKeyRingHelper;
import de.ibh.StrongRoom.DB.DBConn;
import de.ibh.StrongRoom.DB.IEncryptedRow;
import de.ibh.StrongRoom.DB.ISignedRow;
import de.ibh.StrongRoom.DB.PublicKey;
import de.ibh.StrongRoom.DB.SigningKey;
import de.ibh.StrongRoom.Exception.SRBadConfigSignatureException;
import de.ibh.StrongRoom.Exception.SRBadSigningSignatureException;
import de.ibh.StrongRoom.Exception.SRGenericCryptoEngineException;
import de.ibh.StrongRoom.Exception.SRGenericCryptoModelException;
import de.ibh.StrongRoom.Exception.SRMasterKeyNotFoundException;
import de.ibh.StrongRoom.Exception.SRPublicKeyNotFoundException;
import de.ibh.StrongRoom.Exception.SRSecretKeyNotFoundException;
import de.ibh.StrongRoom.Exception.SRSigningKeyBrokenException;
import de.ibh.StrongRoom.Exception.SRSigningKeyMismatchException;
import de.ibh.StrongRoom.Exception.SRSigningKeyNotConfirmedException;
import de.ibh.StrongRoom.Exception.SRSigningSerialException;

public class Crypto extends CryptoBackend {	
	public Crypto(BaseConf conf, boolean verifySignature, IPassphraseRequest ppr) throws SRGenericCryptoModelException {		
		super(ppr);

		String key;

		String keyFPR = conf.getSecretKeyFPR();
        if(keyFPR == null)
        	throw new SRSecretKeyNotFoundException("Property StrongRoom.PGP.SecretKey.keyFPR not set.");

        /* Get key from GnuPG output. */
		if(conf.getSecretKeyKeyMethod() == ESecretKeySource.GNUPG) {
			String gpg = conf.getSecretKeyGPGBin();
			
			if(gpg == null || gpg.isEmpty())
				gpg = "gpg";
			
	        GnuPG.setPath(gpg);
	        GnuPG.updateKeys();

        	key = GnuPG.exportSecretKey(keyFPR);

        	if(key == null || key.isEmpty())
	        	throw new SRSecretKeyNotFoundException("Failed retrieving private key from GnuPG.");
		}
		/* Get key from properties. */
		else {
            key = conf.getSecretKeyData();

            if(key == null || key.isEmpty())
            	throw new SRSecretKeyNotFoundException("Property StrongRoom.PGP.SecretKey.Key not set.");
		}

		/* verify secret key. */
		PGPSecretKeyRing secretKeyRing;
        try {
			secretKeyRing = new PGPSecretKeyRing(
					PGPUtil.getDecoderStream(
							new ByteArrayInputStream(
									key.getBytes()
							)
					)
			);
		} catch (IOException e) {
        	throw new SRSecretKeyNotFoundException("Error reading secret key: " + e.getMessage());
		} catch (PGPException e) {
        	throw new SRSecretKeyNotFoundException("Error importing secret key: " + e.getMessage());
		}
    	
		
		String fpr = formatFPR(secretKeyRing.getSecretKey().getPublicKey().getFingerprint());
		if(!fpr.equalsIgnoreCase(keyFPR))
        	throw new SRSecretKeyNotFoundException("Secret key FPR mismatch (config: " + ("".equals(keyFPR) ? "<none>" : keyFPR) + ", found: " + fpr + ").");
        
		storeSecretKeyRing(ECryptoPGP_KeyType.USER_KEY, secretKeyRing);

		
		/* verify master key */
		PGPPublicKeyRing masterKeyRing;
		try {
			masterKeyRing = new PGPPublicKeyRing(
					PGPUtil.getDecoderStream(
							new ByteArrayInputStream(
									conf.getMasKeyData().getBytes()
							)
					)
			);
		} catch (IOException e) {
        	throw new SRMasterKeyNotFoundException("Error reading master key: " + e.getMessage());
		}

		
		String masterFPR = conf.getMasKeyFPR();

		fpr = formatFPR(masterKeyRing.getPublicKey().getFingerprint());

		if(!fpr.equalsIgnoreCase(masterFPR))
        	throw new SRMasterKeyNotFoundException("Master key FPR mismatch (config: " + masterFPR + ", found: " + fpr + ".");

		storePublicKeyRing(ECryptoPGP_KeyType.MASTER_KEY, masterKeyRing);

		
		/* verify configuration signature */
		if(verifySignature) {
			if(!verifySignature(conf.getSignString().getBytes(), conf.getSignature().getBytes(), secretKeyRing.getPublicKey()))
				throw new SRBadConfigSignatureException("Your configuration has no valid signature.");
		}
	}
	
	public void verifySigningKey(final SigningConf sconf, final SigningKey sk, final IPassphraseRequest ppr, final ISigningKeyVerificationRequest skvr, Object userData) throws SRBadConfigSignatureException, SRBadSigningSignatureException, SRSigningSerialException, SRSigningKeyMismatchException, SRSigningKeyBrokenException, SRSigningKeyNotConfirmedException, SRSecretKeyNotFoundException, SRGenericCryptoEngineException {
        /* verify signing key's master key signature  */
        if(!verifySignedRow(sk))
        	throw new SRBadConfigSignatureException("The signing key DB signature is invalid.");
       
        /* verify signing configuration if it is available */
        if(sconf.getSerial() > 0) {
    		if(!verifySignature(sconf.getSignString().getBytes(), sconf.getSignUKey().getBytes(), sconf.getSignKeyType()))
            	throw new SRBadSigningSignatureException("The signing configuration has no valid signature.");
        }

        /* the serial must be monolithic increasing */
        if(sk.getSerial() < sconf.getSerial())
        	throw new SRSigningSerialException("The signing serial number has been decreased.");

        /* if the serial is the same, the public key must be identical */
        if(sk.getSerial() == sconf.getSerial()) {
        	if(!sk.getPubKey().equals(sconf.getPubKey()))
            	throw new SRSigningKeyMismatchException("The signing key has been changed.");
        }
        else {
        	PGPPublicKeyRing pkr; 
        	try {
				pkr = new PGPPublicKeyRing(
						PGPUtil.getDecoderStream(
								new ByteArrayInputStream(
										sk.getPubKey().getBytes()
								)
						)
				);
			} catch (IOException e1) {
				throw new SRSigningKeyBrokenException("The signing key is invalid.");
			}
        	
			if(!skvr.verifySigningKey(sconf.getSerial(), sconf.getSignString(), sk.getSerial(), formatFPR(pkr.getPublicKey().getFingerprint()), userData))
				throw new SRSigningKeyNotConfirmedException("You did not confirm the current signing key.");
			
			sconf.setSerial(sk.getSerial());
			sconf.setPubKey(sk.getPubKey());

		    byte [] newSig = createSignature(ECryptoPGP_KeyType.USER_KEY, sconf.getSignString().getBytes(), null, userData);
		    
            if(newSig != null) {
                sconf.setSignUKey(Misc.b2s(newSig));
                try {
					sconf.saveConfig();
					
					return;
				} catch (IOException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
            }

            throw new SRSigningKeyNotConfirmedException("You did not sign the current signing key.");
        }
        
        /* store public signing key */
        try {
			storePublicKeyRing(ECryptoPGP_KeyType.SIGNING_KEY, new PGPPublicKeyRing(
					PGPUtil.getDecoderStream(
							new ByteArrayInputStream(
									sk.getPubKey().getBytes()
							)
					)
			));
		} catch (IOException e) {
			/* This should never ever happen! */
			e.printStackTrace();

			assert(false);
		}
	}
	
	public boolean verifySignedRow(ISignedRow sr) {
		String s[] = sr.getSignStrings();
		String data = "<|";
		for(int i = 0; i < s.length; i++) {
			data += s[i] + "|";
		}
		data += ">";

		Map<ECryptoPGP_KeyType, String> sigmap = sr.getSignKeyMap();
		for(Iterator<ECryptoPGP_KeyType> it = sigmap.keySet().iterator(); it.hasNext(); ) {
			ECryptoPGP_KeyType kt = it.next();
			String sig = sr.getSignKeyMap().get(kt);

			if(!verifySignature(data.replaceAll("\r", "").getBytes(), sig.getBytes(), kt))
				return false;
		}
		
		return true;
	}

	public static String buildSignString(String s[]) {		
		String sign = "<|";
		
		for(int i = 0; i < s.length; i++ ) {
			sign += s[i] + "|";
		}
		
		return sign + ">";
	}
	
	public static String getInfo() {
	    return getBouncyCastleVersion();
	}
    
    public void encryptRow(IEncryptedRow er, String data, IRetrievePublicKeyRingHelper h, Object userData) throws SRPublicKeyNotFoundException, SRGenericCryptoEngineException {
    	Set<ECryptoPGP_KeyType> ks = er.getEncKeySet();
    	ECryptoPGP_KeyType ka[] = new ECryptoPGP_KeyType[ks.size()];

    	int i = 0;
    	for(Iterator<ECryptoPGP_KeyType> it = ks.iterator(); it.hasNext(); ) {
    		ka[i++] = it.next();
    	}
    	
    	er.setEncryptedData(Misc.b2s(encrypt(ka, data.getBytes(Charset.forName("UTF-8")), h, userData)));
    }
    
    public void signRow(ISignedRow sr, IRetrieveSecretKeyRingHelper h, Object userData) throws SRSigningKeyNotConfirmedException, SRSecretKeyNotFoundException, SRGenericCryptoEngineException {
    	String t = buildSignString(sr.getSignStrings());

    	for(Iterator<ECryptoPGP_KeyType> it = sr.getSignKeyMap().keySet().iterator(); it.hasNext(); ) {
    		ECryptoPGP_KeyType kt = it.next();
    		
        	byte sig[] = createSignature(kt, t.getBytes(), h, userData);
    	    
            if(sig == null)
            	throw new SRSigningKeyNotConfirmedException("You did not confirmed the signing of the new crypto entry.");

            sr.setSigningData(kt, Misc.b2s(sig));
    	}
    }
    
    @SuppressWarnings("unchecked")
	public Map<String, String> getUIDMap(DBConn db) {
    	List<PublicKey> pubKeys = db.getPublicKeys();
    	Map<String, String> map = new HashMap<String, String>();
    	
    	for(Iterator<PublicKey> it = pubKeys.iterator(); it.hasNext(); ) {
    		PublicKey pk = it.next();
    		String uid = "-";
  
    		PGPPublicKeyRing pkr;
            try {
            	pkr = new PGPPublicKeyRing(
    					PGPUtil.getDecoderStream(
    							new ByteArrayInputStream(
    									pk.getPubKey().getBytes()
    							)
    					)
    			);

    			Iterator<String> i = (Iterator<String>) pkr.getPublicKey().getUserIDs();
    		    if(i.hasNext())
    			    uid = i.next();
    		} catch (Exception e) {
    		}

    		map.put(pk.getFPR(), uid);
    	}
    	
    	return map;
    }
}
