/*
 * Copyright: (c) 2004-2006 Mayo Foundation for Medical Education and
 * Research (MFMER).  All rights reserved.  MAYO, MAYO CLINIC, and the
 * triple-shield Mayo logo are trademarks and service marks of MFMER.
 *
 * Except as contained in the copyright notice above, the trade names,
 * trademarks, service marks, or product names of the copyright holder shall
 * not be used in advertising, promotion or otherwise in connection with
 * this Software without prior written authorization of the copyright holder.
 *
 * Licensed under the Eclipse Public License, Version 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.eclipse.org/legal/epl-v10.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.ibh.StrongRoom.Client.SWT;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.StyleRange;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.widgets.Display;

/**
 * This class makes it easier to display formatted text in a SWT StyledText. You can mark up your text with
 * tags like <b> </b> - basic html tags - and it will create the necessary style ranges for the StyledText
 * object.
 *
 * Supported Tags: (pre is not supported - it is just here so you can read this from the javadocs)
 *
 * <pre>
 * <b> - bold
 * <i> - italic
 * <u> - underline
 * <so> - strikeout
 * <red> - red font color
 * <green> - green font color
 * <blue> - blue font color
 * </pre>
 *
 * @author <A HREF="mailto:armbrust.daniel@mayo.edu">Dan Armbrust</A>
 * @version subversion $Revision: 1.1.2.1 $ checked in on $Date: 2007-05-16 01:10:11 $
 */
public class TextContent {

    protected String content;
    /**
     * @return Returns the content.
     */
    public String getContent() {
        return content;
    }
    /**
     * @param content The content to set.
     */
    public void setContent(String content) {
        this.content = content;
    }

    /**
     * 
     */
    public TextContent(Display display) {
        red = display.getSystemColor(SWT.COLOR_RED);
        green = display.getSystemColor(SWT.COLOR_GREEN);
        blue = display.getSystemColor(SWT.COLOR_BLUE);
        black = display.getSystemColor(SWT.COLOR_BLACK);
        white = display.getSystemColor(SWT.COLOR_WHITE);
    }


    protected int nesting = 0;
        
    protected List<StyleRange> styleRanges = new ArrayList<StyleRange>();

    boolean inItalics = false, inBold = false, inUnderline = false, inStrikeout = false;
    boolean inRed = false, inBlue = false, inGreen = false;

    protected Color black, white, red, green, blue;

    protected StyleRange curRange;
    protected int rangeStyle = SWT.NORMAL;

    protected void startRange(int posn) {
        if (curRange != null) {
            endRange(posn);
        }
        curRange = new StyleRange();
        curRange.start = posn;
    }
    protected void endRange(int posn) {
        if (curRange != null) {
            curRange.length = posn - curRange.start;
            if (inItalics) {
                rangeStyle |= SWT.ITALIC;
            }
            else {
                rangeStyle &= ~SWT.ITALIC;
            }
            if (inBold) {
                rangeStyle |= SWT.BOLD;
            }
            else {
                rangeStyle &= ~SWT.BOLD;
            }
            if (inUnderline) {
                rangeStyle |= SWT.SINGLE;
            }
            else {
                rangeStyle &= ~SWT.SINGLE;
            }            
            curRange.fontStyle = rangeStyle;
            curRange.underline = inUnderline;
            curRange.strikeout = inStrikeout;
            if      (inRed) {
                curRange.foreground = red;
            }
            else if (inGreen) {
                curRange.foreground = green;
            }
            else if (inBlue) {
                curRange.foreground = blue;
            }
            else {
                curRange.foreground = black;
            }
            if (curRange.length > 0) {
                styleRanges.add(curRange);
            }
        }
        curRange = null;
        if (nesting > 0) {
			startRange(posn);
		}
    }
    
    
    public String toFormatedText(StyleRange [] styleRanges) {
        if(styleRanges.length == 0) return content;

        StringBuffer sb = new StringBuffer(content);
        for(int i = 0; i < sb.toString().length(); i++) {
            i = sb.toString().indexOf("<", i);

            if(i == -1)
                break;
            else
                sb.insert(i+1, ">");

            for(StyleRange sr : styleRanges) {
                if(sr.start > i)
                    sr.start++;
                else if(sr.start + sr.length > i)
                    sr.length++;
            }
        }
        String content = sb.toString();
        sb = new StringBuffer(content);

        for(StyleRange sr : styleRanges) {
            StringBuffer startStyle = new StringBuffer();
            StringBuffer endStyle = new StringBuffer();
            if((sr.fontStyle & SWT.BOLD) == SWT.BOLD) {
                startStyle.append("<b>");
                endStyle.append("</b>");                
            }
            if((sr.fontStyle & SWT.ITALIC) == SWT.ITALIC) {
                startStyle.append("<i>");
                endStyle.insert(0, "</i>");
            }
            if((sr.fontStyle & SWT.SINGLE) == SWT.SINGLE) {
                startStyle.append("<u>");
                endStyle.insert(0, "</u>");
            }
            if(sr.foreground == red) {
                startStyle.append("<red>");
                endStyle.insert(0, "</red>");
            }
            if(sr.foreground == green) {
                startStyle.append("<green>");
                endStyle.insert(0, "</green>");
            }
            if(sr.foreground == blue) {
                startStyle.append("<blue>");
                endStyle.insert(0, "</blue>");
            }            

            int offset = Math.abs(content.length() - sb.length());
            sb.insert(sr.start + sr.length + offset, endStyle);
            sb.insert(sr.start + offset, startStyle);            
        }
        
        return sb.toString();
    }

    public String toPlainText() {
        int posn = 0;
        StringBuffer sb = new StringBuffer();
        
        for (int i = 0; i < content.length(); ) {
            char c = content.charAt(i);
            if (c == '<') {
                if      (content.length() >= i + 2 && "<>".equals(content.substring(i, i + 2))) {
                    sb.append('<');
                    i += 2;
                    nesting++;
                    posn++;
                }
                else if (content.length() >= i + 3 && "<i>".equals(content.substring(i, i + 3))) {
                    startRange(posn);
                    inItalics = true;
                    i += 3;
                    nesting++;
                }
                else if (content.length() >= i + 3 && "<b>".equals(content.substring(i, i + 3))) {
                    startRange(posn);
                    inBold = true;
                    i += 3;
                    nesting++;
                }
                else if (content.length() >= i + 3 && "<u>".equals(content.substring(i, i + 3))) {
                    startRange(posn);
                    inUnderline = true;
                    i += 3;
                    nesting++;
                }
                else if (content.length() >= i + 4 && "<so>".equals(content.substring(i, i + 4))) {
                    startRange(posn);
                    inStrikeout = true;
                    i += 4;
                    nesting++;
                }
                else if (content.length() >= i + 5 && "<red>".equals(content.substring(i, i + 5))) {
                    startRange(posn);
                    inRed = true;
                    i += 5;
                    nesting++;
                }
                else if (content.length() >= i + 7 && "<green>".equals(content.substring(i, i + 7))) {
                    startRange(posn);
                    inGreen = true;
                    i += 7;
                    nesting++;
                }
                else if (content.length() >= i + 6 && "<blue>".equals(content.substring(i, i + 6))) {
                    startRange(posn);
                    inBlue = true;
                    i += 6;
                    nesting++;
                }
                else if (content.length() >= i + 4 && "</i>".equals(content.substring(i, i + 4))) {
                    endRange(posn);
                    inItalics = false;
                    i += 4;
                    nesting--;
                }
                else if (content.length() >= i + 4 && "</b>".equals(content.substring(i, i + 4))) {
                    endRange(posn);
                    inBold = false;
                    i += 4;
                    nesting--;
                }
                else if (content.length() >= i + 4 && "</u>".equals(content.substring(i, i + 4))) {
                    endRange(posn);
                    inUnderline = false;
                    i += 4;
                    nesting--;
                }
                else if (content.length() >= i + 5 && "</so>".equals(content.substring(i, i + 5))) {
                    endRange(posn);
                    inStrikeout = false;
                    i += 5;
                    nesting--;
                }
                else if (content.length() >= i + 6 && "</red>".equals(content.substring(i, i + 6))) {
                    endRange(posn);
                    inRed = false;
                    i += 6;
                    nesting--;
                }
                else if (content.length() >= i + 8 && "</green>".equals(content.substring(i, i + 8))) {
                    endRange(posn);
                    inGreen = false;
                    i += 8;
                    nesting--;
                }
                else if (content.length() >= i + 7 && "</blue>".equals(content.substring(i, i + 7))) {
                    endRange(posn);
                    inBlue = false;
                    i += 7;
                    nesting--;
                }
                else {
                    while (true) {
                        if((c = content.charAt(++i)) == '>') {
                            break;
                        }
                    }
                }
            }
            else {
                sb.append(c);
                i++;
                posn++;
            }
        }
        endRange(posn);
        return sb.toString();
    }
    

    public StyleRange[] getStyleRanges() {
        return (StyleRange[])styleRanges.toArray(new StyleRange[styleRanges.size()]);    
    }
}