package de.ibh.StrongRoom.Crypto;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.security.SecureRandom;
import java.security.Security;
import java.security.SignatureException;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.concurrent.atomic.AtomicReference;

import org.bouncycastle.bcpg.ArmoredOutputStream;
import org.bouncycastle.bcpg.BCPGOutputStream;
import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.bouncycastle.openpgp.PGPCompressedData;
import org.bouncycastle.openpgp.PGPCompressedDataGenerator;
import org.bouncycastle.openpgp.PGPEncryptedData;
import org.bouncycastle.openpgp.PGPEncryptedDataGenerator;
import org.bouncycastle.openpgp.PGPEncryptedDataList;
import org.bouncycastle.openpgp.PGPException;
import org.bouncycastle.openpgp.PGPLiteralData;
import org.bouncycastle.openpgp.PGPLiteralDataGenerator;
import org.bouncycastle.openpgp.PGPObjectFactory;
import org.bouncycastle.openpgp.PGPPrivateKey;
import org.bouncycastle.openpgp.PGPPublicKey;
import org.bouncycastle.openpgp.PGPPublicKeyEncryptedData;
import org.bouncycastle.openpgp.PGPPublicKeyRing;
import org.bouncycastle.openpgp.PGPSecretKey;
import org.bouncycastle.openpgp.PGPSecretKeyRing;
import org.bouncycastle.openpgp.PGPSignature;
import org.bouncycastle.openpgp.PGPSignatureGenerator;
import org.bouncycastle.openpgp.PGPSignatureList;
import org.bouncycastle.openpgp.PGPUtil;

import de.ibh.StrongRoom.DB.IEncryptedRow;
import de.ibh.StrongRoom.Exception.SRGenericCryptoEngineException;
import de.ibh.StrongRoom.Exception.SRPublicKeyNotFoundException;
import de.ibh.StrongRoom.Exception.SRSecretKeyNotFoundException;
import de.ibh.StrongRoom.Log.DebugMessage;
import de.ibh.StrongRoom.l10n.StrongRoomL10N;


public class CryptoBackend {
	private IPassphraseRequest ppr;

	private Map<ECryptoPGP_KeyType, PGPPublicKeyRing> publicKeyRingMap = new HashMap<ECryptoPGP_KeyType, PGPPublicKeyRing>();
    private Map<ECryptoPGP_KeyType, PGPSecretKeyRing> secretKeyRingMap = new HashMap<ECryptoPGP_KeyType, PGPSecretKeyRing>();
    static DebugMessage                               logger            = DebugMessage.getInstance();
    static StrongRoomL10N                             l10n = StrongRoomL10N.getInstance();

    public enum ECryptoPGP_KeyType {
        MASTER_KEY,
        SIGNING_KEY,
        ROLE_KEY,
        USER_KEY,
    }
    
    public static String getBouncyCastleVersion() {
        return (new BouncyCastleProvider()).getInfo();
    }
    
    
    public CryptoBackend(final IPassphraseRequest ppr) {
    	if(Security.getProvider(BouncyCastleProvider.PROVIDER_NAME) == null) {
    		Security.addProvider(new BouncyCastleProvider());
    	}

    	this.ppr = ppr;
    }    
    
    public static String keyLong2Hex (long keyid) {
        if (Long.toHexString(keyid).length() < 8) return null;
        
        return Long.toHexString(keyid).toUpperCase().substring(8);
    }
    
    public void storeSecretKeyRing(ECryptoPGP_KeyType mIndex, PGPSecretKeyRing secretKeyRing) {
    	secretKeyRingMap.put(mIndex, secretKeyRing);
    }
    
    public void storePublicKeyRing(ECryptoPGP_KeyType mIndex, PGPPublicKeyRing publicKeyRing) {
    	publicKeyRingMap.put(mIndex, publicKeyRing);
    }
    
    @SuppressWarnings("unchecked")
    public PGPPublicKey retrievePublicEncryptionKey(ECryptoPGP_KeyType mIndex) {
        PGPPublicKeyRing pkr = publicKeyRingMap.get(mIndex);

        if(pkr != null) {
            for(Iterator<PGPPublicKey> it = (Iterator<PGPPublicKey>) pkr.getPublicKeys(); it.hasNext(); ) {
                PGPPublicKey pk = it.next();
                if(pk.isEncryptionKey())
                    return pk;
            }
        }

        PGPSecretKeyRing skr = secretKeyRingMap.get(mIndex);
        
        if(skr != null) {
            for(Iterator<PGPSecretKey> it = (Iterator<PGPSecretKey>) skr.getSecretKeys(); it.hasNext(); ) {
                PGPPublicKey pk = it.next().getPublicKey();
                if(pk.isEncryptionKey())
                    return pk;
            }
        }
        
        return null;
    }
    
    public PGPPublicKey retrievePublicSigningKey(ECryptoPGP_KeyType mIndex) {
        PGPPublicKeyRing pkr = publicKeyRingMap.get(mIndex);

        if(pkr != null)
            return pkr.getPublicKey();

        PGPSecretKeyRing skr = secretKeyRingMap.get(mIndex);
        
        if(skr != null)
            return skr.getPublicKey();
        
        return null;
    }
    
    
    public String getFPR(ECryptoPGP_KeyType mIndex) {
        PGPPublicKey pubkey = retrievePublicSigningKey(mIndex);
        
        if(pubkey == null)
        	return null;

        return formatFPR(pubkey.getFingerprint());
    }
    
	public static String formatFPR(byte in[]) {
		String out = "";
		
		for(int i = 0; i < in.length; i++) {
			out += String.format("%02X", in[i]);
			if(i % 2 == 1)
				out += " ";

			if(i == 9)
				out += " ";
		}

		return out.trim();
	}

	public static String formatFPR(String fingerprint) {
		if(fingerprint == null || fingerprint.length() != 40)
			return "";

		String ret = "";
		
		for(int i = 0; i < 10; i++) {
			ret += fingerprint.substring(i*4, (i + 1) *4);

			if(i == 4)
				ret += "  ";
			else if(i != 9)
				ret += " ";
		}
		
		return ret;
	}
	    
    public long getSecretKeyID(String mIndex) {
        PGPSecretKey secretKey = null;
        
        try {
            secretKey = secretKeyRingMap.get(mIndex).getSecretKey();
        } catch (Exception e) {
            // TODO: Error handling
            return 0;
        }
        
        return secretKey == null ? 0 : secretKey.getKeyID();
    }
    
    
    public byte[] decrypt(ECryptoPGP_KeyType mIndex, IEncryptedRow data, Object userData) throws SRSecretKeyNotFoundException, SRGenericCryptoEngineException {
        return decrypt(mIndex, data.getEncryptedData().getBytes(), ppr, null, userData);
    }

    
	@SuppressWarnings("unchecked")
	public byte[] decrypt(String privKey, byte[] encData, Object userData) throws SRSecretKeyNotFoundException, SRGenericCryptoEngineException {
	    Map<Long, PGPPrivateKey> privateKeys = new HashMap<Long, PGPPrivateKey>();
        
		PGPSecretKeyRing secretKeyRing;
		try {
			secretKeyRing = new PGPSecretKeyRing(
					PGPUtil.getDecoderStream(
							new ByteArrayInputStream(
									privKey.getBytes()
							)
					)
			);
		} catch (IOException e) {
			throw new SRGenericCryptoEngineException(e);
		} catch (PGPException e) {
			throw new SRGenericCryptoEngineException(e);
		}
        
        
        char [] passPhrase = null;
        if(secretKeyRing.getSecretKey().getKeyEncryptionAlgorithm() != PGPEncryptedData.NULL) {
    		String pp = ppr.getPassphrase(
            		l10n._("Unlock the secret key for decryption"),
            		false,
            		secretKeyRing.getPublicKey(),
            		userData
            );
    		
    		/* Cancel button */
    		if(pp == null)
    			return null;
    		
    		passPhrase = pp.toCharArray();
        }

        try {
        	for(Iterator<PGPSecretKey> i = secretKeyRing.getSecretKeys(); i.hasNext(); ) {
        		PGPSecretKey s = i.next();
        		
        		if(s.getPublicKey().isEncryptionKey())
        		    privateKeys.put(s.extractPrivateKey(passPhrase, "BC").getKeyID(), s.extractPrivateKey(passPhrase, "BC"));
        	}
        } catch (PGPException pgpException) {
            ppr.clearInput();
			throw new SRGenericCryptoEngineException(pgpException);
        } catch (NoSuchProviderException nspException) {
            ppr.clearInput();
			throw new SRGenericCryptoEngineException(nspException);
        }
        finally {
            /* forget the passphrase */
            passPhrase = null;
            System.gc();
        }
        
        return decrypt(encData, privateKeys);  
    }
    
    @SuppressWarnings("unchecked")
	public byte[] decrypt(final ECryptoPGP_KeyType mIndex, final byte[] encData, final IPassphraseRequest pr, final IRunnable ira, final Object userData) throws SRSecretKeyNotFoundException, SRGenericCryptoEngineException {
        final PGPSecretKeyRing  secretKeyRing  = secretKeyRingMap.get(mIndex);
        
        char [] passPhrase = null;
        if(secretKeyRing.getSecretKey().getKeyEncryptionAlgorithm() != PGPEncryptedData.NULL) {
        	final AtomicReference<String> pp = new AtomicReference<String>(); 
        	if(ira != null)
        		ira.run(new Runnable() {
					
					@Override
					public void run() {
			    		pp.set(pr.getPassphrase(
			    				l10n._("Unlock the secret key for decryption"),
			            		false,
			            		secretKeyRing.getPublicKey(),
			            		userData));
					}
				});
        	else 
	    		pp.set(pr.getPassphrase(
	            		l10n._("Unlock the secret key for decryption"),
	            		false,
	            		secretKeyRing.getPublicKey(),
	            		userData));
    		
    		/* Cancel button */
    		if(pp.get() == null)
    			return null;
    		
    		passPhrase = pp.get().toCharArray();
        }
        
        Map<Long, PGPPrivateKey> privateKeys = new HashMap<Long, PGPPrivateKey>();
        try {
        	for(Iterator<PGPSecretKey> i = secretKeyRing.getSecretKeys(); i.hasNext(); ) {
        		PGPSecretKey s = i.next();
        		
        		if(s.getPublicKey().isEncryptionKey())
        		    privateKeys.put(s.extractPrivateKey(passPhrase, "BC").getKeyID(), s.extractPrivateKey(passPhrase, "BC"));
        	}
        } catch (PGPException pgpException) {
            ppr.clearInput();
			throw new SRGenericCryptoEngineException(pgpException);
        } catch (NoSuchProviderException nspException) {
            ppr.clearInput();
			throw new SRGenericCryptoEngineException(nspException);
        }
        finally {
            /* forget the passphrase */
            passPhrase = null;
            System.gc();
        }
        
        return decrypt(encData, privateKeys);  
    }
    
    
    public byte[] createSignature(ECryptoPGP_KeyType kt, byte[] data, IRetrieveSecretKeyRingHelper h, Object userData) throws SRSecretKeyNotFoundException, SRGenericCryptoEngineException {
    	PGPSecretKeyRing skr = null;

    	if(secretKeyRingMap.containsKey(kt))
            skr = secretKeyRingMap.get(kt);
    	else {
    		String key = h.getSecretKeyring(kt, userData);
    		
    		if(key == null)
    			throw new SRSecretKeyNotFoundException(l10n._("Could not get unexpected secret key ") + kt.name() + ".");

            try {
    			skr = new PGPSecretKeyRing(
    					PGPUtil.getDecoderStream(
    							new ByteArrayInputStream(
    									key.getBytes()
    							)
    					)
    			);
    		} catch (IOException e) {
    			throw new SRGenericCryptoEngineException(e);
    		} catch (PGPException e) {
    			throw new SRGenericCryptoEngineException(e);
    		}
    	}
    	
        return createSignature(data, skr.getSecretKey(), userData);
    }
    
    
    @SuppressWarnings("unchecked")
    public byte[] encrypt(ECryptoPGP_KeyType mIndex[], byte data[], IRetrievePublicKeyRingHelper h, Object userData) throws SRPublicKeyNotFoundException, SRGenericCryptoEngineException {
        PGPPublicKey publicKeys[] = new PGPPublicKey[mIndex.length];
       
        for(int i = 0; i < mIndex.length; i++) {
            publicKeys[i] = retrievePublicEncryptionKey(mIndex[i]);

        	if(publicKeys[i] == null && h != null){
	    		String key = h.getPublicKeyring(mIndex[i], userData);
	    		
	    		if(key == null)
	    			throw new SRPublicKeyNotFoundException(l10n._("Could not get unexpected public key ") + mIndex[i].name() + ".");
	
	            try {
	            	 PGPPublicKeyRing pkr = new PGPPublicKeyRing(
	    					PGPUtil.getDecoderStream(
	    							new ByteArrayInputStream(
	    									key.getBytes()
	    							)
	    					)
	            	 );
	            	 
	            	 for(Iterator<PGPPublicKey> it = (Iterator<PGPPublicKey>)pkr.getPublicKeys(); it.hasNext(); ) {
	            	     PGPPublicKey pk = it.next();
	            	     
	            	     if(pk.isEncryptionKey()) {
	            	         publicKeys[i] = pk;
	            	         break;
	            	     }
	            	 }
	            	 
	    		} catch (IOException e) {
	                throw new SRPublicKeyNotFoundException(l10n._("Error reading public key ") + mIndex[i].name() + ": " + e.getMessage());
				}
        	}
    		
    		if(publicKeys[i] == null)
    			throw new SRPublicKeyNotFoundException(l10n._("The public key ") + mIndex[i].name() + l10n._(" is not known!"));
        }

        return encrypt(data, publicKeys);
    }
  

    private static byte[] decrypt(byte[] encData, Map<Long, PGPPrivateKey> privKeys) throws SRSecretKeyNotFoundException, SRGenericCryptoEngineException {
        byte[] r = null;
        
        try {
            if(privKeys.size() == 0)
                throw new SRSecretKeyNotFoundException(l10n._("Got an empty secret key."));
            
            InputStream in = new ByteArrayInputStream(encData);
   
            PGPObjectFactory pgpFact = new PGPObjectFactory(PGPUtil.getDecoderStream(in));
            PGPEncryptedDataList enc;
            PGPPublicKeyEncryptedData pub = null;
            Object obj = pgpFact.nextObject();
            
            if(obj instanceof  PGPEncryptedDataList)
                enc = (PGPEncryptedDataList) obj;
            else
                enc = (PGPEncryptedDataList) pgpFact.nextObject();
            
            PGPPrivateKey privKey = null;
            for(int i = 0; i < enc.size(); i++) {
            	PGPPublicKeyEncryptedData _pub = (PGPPublicKeyEncryptedData) enc.get(i);
            	for(Iterator<Long> it = privKeys.keySet().iterator(); it.hasNext();) {
            	    Long privKeyID = it.next();
                    if(_pub.getKeyID() == privKeyID) {
                    	pub = _pub;
                    	privKey = privKeys.get(privKeyID);
                        break;
                    }
            	}
            	if(pub != null)
            	    break;
            }

            if(pub == null)
            	throw new SRSecretKeyNotFoundException(l10n._("Can not decrypt data, because no secret key is available!"));
            
            InputStream clear = pub.getDataStream(privKey, "BC");
    
            pgpFact = new PGPObjectFactory(clear);
    
            PGPCompressedData cData = (PGPCompressedData) pgpFact
                    .nextObject();
    
            pgpFact = new PGPObjectFactory(cData.getDataStream());
    
            PGPLiteralData litData = null;
            while((obj = pgpFact.nextObject()) != null) {
            	if(obj instanceof PGPLiteralData) {
            		litData = (PGPLiteralData) obj;
            		break;
            	}
            }
            
            if(litData == null)
            	throw new SRSecretKeyNotFoundException(l10n._("Can not decrypt data, because litData not found!"));
            
            InputStream plain = litData.getInputStream();
    
            ByteArrayOutputStream output = new ByteArrayOutputStream();
            int c;
    
            while ((c = plain.read()) >= 0) {
                output.write(c);
            }
    
            r = output.toByteArray();
            output.close();
        } catch (PGPException pgpException) {
            if(pgpException.getUnderlyingException() != null  
                    && pgpException.getUnderlyingException() instanceof InvalidKeyException) {
                logger.debug(l10n._("Failed to decrypt: ") + pgpException.getMessage());
            }

            throw new SRGenericCryptoEngineException(pgpException);
        } catch (IOException ioException) {
            throw new SRGenericCryptoEngineException(ioException);
        } catch (NoSuchProviderException nspException) {
            throw new SRGenericCryptoEngineException(nspException);
        }

        return r;
    }
    
    private static byte[] encrypt(byte[] data, PGPPublicKey pKeys[]) throws SRGenericCryptoEngineException {
        byte [] r = null;
        
        try {
            ByteArrayOutputStream encOut = new ByteArrayOutputStream();
    
            OutputStream out = encOut;
            out = new ArmoredOutputStream(out);
    
            ByteArrayOutputStream bOut = new ByteArrayOutputStream();
    
            PGPCompressedDataGenerator comData = new PGPCompressedDataGenerator(
                    PGPCompressedDataGenerator.ZLIB);
            OutputStream cos = comData.open(bOut);
            PGPLiteralDataGenerator lData = new PGPLiteralDataGenerator();
    
            OutputStream pOut = lData.open(cos, PGPLiteralData.BINARY, PGPLiteralData.CONSOLE,
                                            data.length, new Date());
            pOut.write(data);
    
            lData.close();
            comData.close();
    
         // TODO: Configurable EncryptionAlgorithm.
            PGPEncryptedDataGenerator cPk = new PGPEncryptedDataGenerator(
                    PGPEncryptedData.AES_256, true, // true: Integrity check
                    new SecureRandom(), "BC");
    
            for(int i = 0; i < pKeys.length; i++) {
            	cPk.addMethod(pKeys[i]);
            }
    
            byte[] bytes = bOut.toByteArray();
    
            OutputStream cOut = cPk.open(out, bytes.length);
            cOut.write(bytes);
    
            cOut.close();
            out.close();
    
            r = encOut.toByteArray();
        } catch (PGPException pgpException) {
            if(pgpException.getUnderlyingException() != null  
                    && pgpException.getUnderlyingException() instanceof InvalidKeyException) {
                logger.debug(l10n._("Failed to decrypt: ") + pgpException.getMessage());                
                pgpException.printStackTrace();
            }            
            
            throw new SRGenericCryptoEngineException(pgpException);
        } catch (IOException ioException) {
            throw new SRGenericCryptoEngineException(ioException);
        } catch (NoSuchProviderException nspException) {
            throw new SRGenericCryptoEngineException(nspException);
        }
        
        return r;
    }
    
    
    private byte [] createSignature(byte [] data, PGPSecretKey secretKey, Object userData) throws SRGenericCryptoEngineException {
        byte [] r = null;
        PGPPrivateKey privateKey = null;
        
        try {
            // TODO: Configurable Encryption- and HashAlgorithm.
            final PGPSignatureGenerator sPk = new PGPSignatureGenerator(secretKey.getPublicKey().getAlgorithm(), 
                    PGPUtil.SHA256, "BC");
            
            char [] passPhrase = null;
            if(secretKey.getKeyEncryptionAlgorithm() != PGPEncryptedData.NULL) {
            	String pp = ppr.getPassphrase(
            			l10n._("Unlock the secret key to create a signature"),
            			true,
            			secretKey.getPublicKey(),
            			userData);

            	if(pp == null)
            		return null;
            	
            	passPhrase = pp.toCharArray();
            }
            privateKey = secretKey.extractPrivateKey(passPhrase, "BC");

            /* forget the passphrase */
            passPhrase = null;
            System.gc();
            
            sPk.initSign(PGPSignature.BINARY_DOCUMENT, privateKey);

            // Data generator
            ByteArrayOutputStream sigOut = new ByteArrayOutputStream();
            OutputStream out = sigOut;
            out = new ArmoredOutputStream(out);
            
            sPk.update(data);
            
            BCPGOutputStream bOut = new BCPGOutputStream(out);

            sPk.generate().encode(bOut);
            out.close();

            r = sigOut.toByteArray();
        } catch (NoSuchAlgorithmException e) {
            ppr.clearInput();
			throw new SRGenericCryptoEngineException(e);
        } catch (NoSuchProviderException e) {
            ppr.clearInput();
			throw new SRGenericCryptoEngineException(e);
        } catch (PGPException e) {
            ppr.clearInput();
			throw new SRGenericCryptoEngineException(e);
        } catch (SignatureException e) {
            ppr.clearInput();
			throw new SRGenericCryptoEngineException(e);
        } catch (IOException e) {
            ppr.clearInput();
			throw new SRGenericCryptoEngineException(e);
        }
        
        return r;
    }
    
    public boolean verifySignature(byte [] data, byte [] sig, ECryptoPGP_KeyType kt) {
    	PGPPublicKey pKey = null;
    	
    	if(publicKeyRingMap.containsKey(kt))
    		pKey = publicKeyRingMap.get(kt).getPublicKey();
    	else if(secretKeyRingMap.containsKey(kt))
    		pKey = secretKeyRingMap.get(kt).getPublicKey();
    	
    	if(pKey == null)
    		return false;
    	
    	return verifySignature(data, sig, pKey);
    }
    
    public static boolean verifySignature(byte [] data, byte [] sig, PGPPublicKey pKey) {
        boolean r = false;

        try {
            InputStream in0 = new ByteArrayInputStream(sig);
            InputStream in = PGPUtil.getDecoderStream(in0);
            PGPObjectFactory pgpFact = new PGPObjectFactory(in);
            
            Object obj = pgpFact.nextObject();
            if((obj instanceof PGPSignatureList) == false) {
                return false;
            } 
            // TODO: check if obj instanceof PGPCompressedData
            
            PGPSignatureList list = (PGPSignatureList)obj;
            PGPSignature pgpSig = null;
            for(int i=0; i<list.size(); i++) {
                if(list.get(i).getKeyID() == pKey.getKeyID())
                    pgpSig = list.get(i);
            }
            
            if(pgpSig == null) {
                logger.debug(l10n._("No valid signature found for this key (ID: ") + pKey.getKeyID() + ")");
                return false;
            }
            
            pgpSig.initVerify(pKey, "BC");
            pgpSig.update(data);
            r = pgpSig.verify();
        } catch (SignatureException e) {
            logger.debug(e);
        } catch (NoSuchProviderException e) {
            logger.debug(e);
        } catch (PGPException e) {
            logger.debug(e);
        } catch (IOException e) {
            logger.debug(e);
        } 
        
        return r;
    }
}